structure ExprBiNF  = struct

local

open BiNF ExprDD TypeDD 

in

structure ExprInterface = struct

type Xpr      = Expr
type AtomCode = int

datatype 'a iExpr = 
	  iTrivialAtom	
	| iContraAtom
	| iAtom 	of 'a
	| iNot 		of 'a iExpr
	| iAnd 		of 'a iExpr * 'a iExpr
	| iOr  		of 'a iExpr * 'a iExpr


local

val ty_bool = TyCons ("bool",[])


(* we assume the expression has been type-checked *)

fun is_NOT  (sym,ty) = (sym = "~")

val ty_bool_binop 
	= 
	TyCons ("fun", [ty_bool, 
			(TyCons ("fun", [ty_bool, ty_bool]))])

fun is_binop sym1 (sym2,ty) = (sym1 = sym2) andalso (ty = ty_bool_binop)

val ty_NOT = TyCons ("fun", [ty_bool, ty_bool])

val mk_T       = Lit ("T",ty_bool)
val mk_F       = Lit ("F",ty_bool)
fun mk_OR  p q = Binary (("\\/",ty_bool_binop),p,q)
fun mk_AND p q = Binary (("/\\",ty_bool_binop),p,q)
fun mk_IMP p q = Binary (("==>",ty_bool_binop),p,q)
fun mk_EQ  p q = Binary (("=",ty_bool_binop),p,q)
fun mk_NOT p   = Unary  (("~",ty_NOT),p)

(* 
This function rewrite p ==> q to ~p \/ q. The rewrite does not
penetrate quantification and enumeration however, since these are
regarded as atoms.

Keep in mind the restriction that control location label isn't allowed
to occur inside a quantification nor an enumeration.
*)

fun rewrite_imp e = case e of

	  Binary (opr,p,q)	=>

		if      is_binop "/\\" opr then mk_AND (rewrite_imp p) (rewrite_imp q)
		else if is_binop "\\/" opr then mk_OR  (rewrite_imp p) (rewrite_imp q)
		else if is_binop "==>" opr then mk_OR  (mk_NOT (rewrite_imp p)) (rewrite_imp q)
		else if is_binop "="   opr then mk_EQ  (rewrite_imp p) (rewrite_imp q) 
		else e

	| Unary (opr,p)		=>
		if is_NOT opr then mk_NOT (rewrite_imp p) else e

	| _			=> e

fun rewrite_eq e = case e of

	  Binary (opr,p,q)	=>

		if      is_binop "/\\" opr then mk_AND (rewrite_eq p) (rewrite_eq q)
		else if is_binop "\\/" opr then mk_OR  (rewrite_eq p) (rewrite_eq q)
		else if is_binop "==>" opr then mk_IMP (rewrite_eq p) (rewrite_eq q) 
		else if is_binop "="   opr then mk_AND (mk_IMP (rewrite_eq p) (rewrite_eq q))
						     (mk_IMP (rewrite_eq q) (rewrite_eq p))

		else e

	| Unary (opr,p)		=>
		if is_NOT opr then mk_NOT (rewrite_eq p) else e

	| _			=> e


(* This function removes double negation, applies de Morgan, and
removes ~T and ~F. So, negation is minimized. We assume ==> and = have
been removed before this function is called. *)

fun reduceNeg e = case e of

	  Unary (("~",_), Unary (("~",_), f))	
		=> reduceNeg f

	| Unary (("~",_), Binary (("/\\",_),f,g))
		=> mk_OR (reduceNeg (mk_NOT f)) (reduceNeg (mk_NOT g))

	| Unary (("~",_), Binary (("\\/",_),f,g))
		=> mk_AND (reduceNeg (mk_NOT f)) (reduceNeg (mk_NOT g))

	| Unary (("~",_), Lit ("T",TyCons ("bool",[])))
		=> mk_F

	| Unary (("~",_), Lit ("F",TyCons ("bool",[])))
		=> mk_T

	| Binary (("/\\",_),f,g)
		=> mk_AND (reduceNeg f) (reduceNeg g)

	| Binary (("\\/",_),f,g)
		=> mk_OR (reduceNeg f) (reduceNeg g)

	| _	=> e


(* A function to reduce T and F. After this, the expression will be in
a proto normal form. To normalize it, we later convert it to the
interface data type iExpr, then call a normalizer provided by some NF
structure/module. 

We assume we have done reduceNeg first before calling this function.
*)

fun reduceTF e 
	=
	let

	fun tc_reduce e = case e of 


		  Binary (("/\\",_), Lit("F",TyCons ("bool",[])), _)
			=> mk_F

		| Binary (("/\\",_), _, Lit("F",TyCons ("bool",[])))
			=> mk_F

		| Binary (("/\\",_), Lit("T",TyCons ("bool",[])), q)
			=> q
		| Binary (("/\\",_), p, Lit("T",TyCons ("bool",[])))
			=> p

		| Binary (("\\/",_), Lit("T",TyCons ("bool",[])), _)
			=> mk_T

		| Binary (("\\/",_), _, Lit("T",TyCons ("bool",[])))
			=> mk_T

		| Binary (("\\/",_), Lit("F",TyCons ("bool",[])), q)
			=> q
		| Binary (("\\/",_), p, Lit("F",TyCons ("bool",[])))
			=> p

		| _	=> e

	fun worker e = case e of

		  Binary (("/\\",_),f,g)
			=> tc_reduce (mk_AND (worker f) (worker g))

		| Binary (("\\/",_),f,g)
			=> tc_reduce (mk_OR (worker f) (worker g))

		| _	=> e

	in
	worker e
	end

(* We assume the following functions to test atom anc collect atoms
are applied after the expression has been transformed to proto NF.

We also assume that the expression has been type checked.*)

(* The function isAtom doesn't really test if e is an atom. It only does
so if it is called by collectAtoms *)

fun isAtom e = case e of

	  Binary (opr,p,q)
		=>
		not(is_binop "/\\" opr)	andalso
		not(is_binop "\\/" opr)

	| Unary (opr,_)			=> not (is_NOT opr)

	| _				=> true


(* A function to collect atoms. T and F are not included in the
result. Duplicates are removed. The return is an array, which
represents a dictionary of all atoms (except T and F) occuring in an
expression. *)

fun collectAtoms e 
	=
	let
	fun collectAllAtoms e 
		=
		if isAtom e 
		then (if e=mk_T orelse e=mk_F then [] else [e])
		else
	
		case e of
 
		  Binary (_,p,q)	=> collectAllAtoms p @ collectAllAtoms q
		| Unary (_,p)		=> collectAllAtoms p
		| _			=> (* actually not possible *) [e]
	in
	Array.fromList (MoreList.nub (collectAllAtoms e))
	end
	
fun encodeAtom atoms_table e
	=
	let
	val N = Array.length atoms_table
	fun search n 
		=
		if n=N then NONE
		else if Array.sub (atoms_table,n) = e 
			then SOME n
			else search (n+1)
	in
	search 0
	end	

fun decodeAtom atoms_table n
	=
	if n<0 orelse n>= Array.length atoms_table
	then NONE
	else SOME (Array.sub (atoms_table,n))


in


(* This function rewrites an expression to a proto normal form, then
casts to an iExpr value. It also returns a decode function to be used
to translate the resulting normalized iExpr back to Expr value. *)


fun to_iExpr e 
	=
	let
	val e' = (reduceTF o reduceNeg o rewrite_imp o rewrite_eq) e
	val atoms_table = collectAtoms e'
	fun code a = valOf (encodeAtom atoms_table a)

	fun translate e
		=
		if isAtom e 
		then
			(case e of
		  	    Lit ("T", TyCons ("bool",[])) => iTrivialAtom
			  | Lit ("F", TyCons ("bool",[])) => iContraAtom
			  | _				   => iAtom (code e)
			)

		else
 			(case e of

		  	   Binary (opr,p,q)		=>

				if is_binop "/\\" opr
				then iAnd (translate p, translate q)
				else if is_binop "\\/" opr 
				then iOr  (translate p, translate q)
				else (* actually not possible *)
				iAtom (code e)

			| Unary (opr,p)			=>
			
				if is_NOT opr 
				then iNot (translate p)
				else (* actually not possible *)
				iAtom (code e)
			| _				=>
				(* also not possible *)
				iAtom (code e)
			)
	in
	(translate e', decodeAtom atoms_table)
	end


fun from_iExpr iexpr = case iexpr of

	  iTrivialAtom		=> mk_T
	| iContraAtom		=> mk_F
	| iAtom undecoded_atom	=> undecoded_atom
	| iNot p		=> mk_NOT (from_iExpr p)
	| iAnd (p,q)		=> mk_AND (from_iExpr p) (from_iExpr q)
	| iOr (p,q)		=> mk_OR  (from_iExpr p) (from_iExpr q)


end
end


structure AtomCode_STRUCT = struct

type Atom = int

fun compare (a,b) = Int.compare (a,b)
	
end

structure Clause = Clause0_FUNC AtomCode_STRUCT

structure NF = BiNF_FUNC Clause

structure ExprNF 
	= 
	Normalizer_FUNC (structure ExprBus = ExprInterface
			 structure Clause  = Clause
			 structure NF      = NF)


end
end




