{
 open Lexing ExprParser;

 exception LexicalError of string * int * int (* (message, loc1, loc2) *)

 fun lexerError lexbuf s = 
     raise LexicalError (s, getLexemeStart lexbuf, getLexemeEnd lexbuf);

 val commentStart = ref 0;  (* Start of outermost comment being scanned *)
 
 fun commentNotClosed lexbuf =
     raise LexicalError ("Comment not terminated", 
                         !commentStart, getLexemeEnd lexbuf);
     
val commentDepth = ref 0;  (* Current comment nesting *)

 (* Scan keywords as identifiers and use this function to distinguish them. *)
 (* If the set of keywords is large, use an auxiliary hashtable.            *)

fun keyword s =
     case s of
         "T"            => LIT_TRUE
       | "F"            => LIT_FALSE
       | "int"		=> INT
       | "bool"		=> BOOL
       | "max"		=> MAX
       | "min"		=> MIN
       | "div"          => DIV
       | "mod"          => MOD
       | "elem"         => ELEM
       | "IN"           => IN
       | "SET"		=> SET
       | _              => IDENT s

fun special s = case s of
	  "="		=> EQUAL
	| "~"		=> NOT
	| "/\\"		=> AND
	| "\\/"		=> OR
	| "==>"		=> IMPLIES
	| "?"		=> EXISTS
	| "!"		=> FORALL
	| "<"		=> LT
	| "<="		=> LTE
	| ">"		=> GT
	| ">="		=> GTE
	| "+"		=> PLUS
	| "-"		=> MINUS
	| "*"		=> TIMES
	| "^"		=> POWER
	| "/"		=> DIVIDE
	| "++"		=> APPEND
	| "!!"		=> NTH
	| "|_|"		=> UNION
	| "|^|"		=> INTERSECT
	| "<+"		=> INSERT
	| "#"		=> ARRAYACCESS	
	| ">>"		=> FUNAPP
	| "["		=> LSBRAC
	| "]"		=> RSBRAC
	| _             => IDENT s
 }

rule Token = parse
    [` ` `\t` `\n` `\r`]     
			{ Token lexbuf 					}

  | "(*"                { commentStart := getLexemeStart lexbuf;
                          commentDepth := 1; 
                          SkipComment lexbuf; Token lexbuf 		}

  | `-`? [`0`-`9`]+     { LIT_I (getLexeme lexbuf)			}
  
  | [`a`-`z``A`-`Z`][`a`-`z``A`-`Z``0`-`9` `'` `_` `.` ]*
                        
 			{ keyword (getLexeme lexbuf) 			}

  | [`~` `\` `/` `=` `>` `<` `-` `+` `!` `?` `*` `^` `|` `_` `#`]+

			{ special (getLexeme lexbuf) 			}

  | `(`                 { LPAR }
  | `)`                 { RPAR }
  | "-{}-"		{ LIT_EMPTYSET }
  | "-[]-"		{ LIT_EMPTYLIST }
  | "-["		{ LLISTBRAC }
  | "]-"		{ RLISTBRAC }
  | "-{"		{ LSETBRAC }
  | "}-"		{ RSETBRAC }
  | `[`                 { LSBRAC }
  | `]`                 { RSBRAC }
  | `{`                 { LACOL }
  | `}`                 { RACOL }
  | `;`                 { SEMI }
  | `,`                 { COMA }
  | "::"		{ CONS }
  | `:`                 { COLON }

  | eof                 { EOF }
  | _                   { lexerError lexbuf "Illegal symbol in input" }

and SkipComment = parse
    "*)"                { commentDepth := !commentDepth - 1;  
                          if !commentDepth = 0 then ()
                          else SkipComment lexbuf 
                        } 
   | "(*"               { commentDepth := !commentDepth + 1; 
                          SkipComment lexbuf }
   | (eof | `\^Z`)      { commentNotClosed lexbuf }
   | _                  { SkipComment lexbuf }
;




