structure ExprLib = struct

(* ============================================================
(*DOC

<MODULE> ExprLib
<AUTHOR> underdarkPrime
<ADDRESS> underdarkPrime@yahoo.com
<DATE> 7 August 2001

<DESC>Provide utilities to work on 'expression' (value of type
ExprDD.Expr). 

DOC*)
============================================================ *)

local 

open TypeDD ExprDD TypeLib exConst MoreList

infix @@

fun replicate n x 
	=
	if n<=0 then [] else x :: replicate (n-1) x

fun fst (x,_) = x

in


fun failMesg loc expr extra =    loc ^":" 
				^ "in Expression  "
				^ ExprPP.toString 30 expr 
				^ "  "
				^ extra ;
 


(* ============================================================
(*DOC

<ITEM>inventNewVarName
<TYPE>inventNewVarName : int -> string
<CALL>inventNewVarName i

<DESC>Converts i to a string representing a variable name. Each
(non-negative) integer i will be translated to a unique name. This
used on places where fresh variable names need to be generated.

DOC*)
============================================================ *)

fun inventNewVarName n 
	=
	let
	val a = String.str (chr(ord #"a" + n mod 26))
		^  	
		(if n>=26 then Int.toString (n div 26) else "")
	in
	TyVar a
	end

(* ============================================================
(*DOC

<ITEM>freshenTyVars
<TYPE>freshenTyVars : int -> TypeDD.Type -> (int * TypeDD.Type)
<CALL>freshenTyVars i ty

<DESC>Renames all type variables occuring in ty with new names, using
i as the base counter to generate new names. It also returns the next
free integer (to be used to generate the next fresh type variable).

<ITEM>freshenTyVars0
<TYPE>freshenTyVars0 : int -> string list -> (string * TypeDD.Type) list
<CALL>freshenTyVars0 i vars

<DESC> When given a list of sting representing variable names, it the
type substitution that refreshes the type variables in vars. The
integer i is used as the base to generate fresh names.

DOC*)
============================================================ *)


fun freshenTyVars0 n []      = []
  | freshenTyVars0 n (a::vs) = (a, inventNewVarName n) :: freshenTyVars0 (n+1) vs

fun freshenTyVars n ty
	=
	let
	val tvars = tv ty
	val subst	= freshenTyVars0 n tvars
	val n'		= n + length tvars
	in
	(n', tySubst subst ty)
	end	


(* ============================================================
(*DOC

<ITEM>tySpec
<TYPE>tySpec : (string * TypeDD.Type) list -> ExprDD.Expr -> ExprDD.Expr
<CALL>tySpec subst expr

<DESC>Applies the type substitution <tt>subst</tt> to the expression.

DOC*)
============================================================ *)

fun tySpec subst expr = case expr of

	  Ident (x,ty) 		=> Ident (x, tySubst subst ty)

	| Lit (x,ty) 		=> Lit   (x, tySubst subst ty)

	| ArrayAccess (a,i) 	=> ArrayAccess (tySpec subst a, tySpec subst i)
			
	| App (f,x) 		=>  App (tySpec subst f, tySpec subst x)

	| Unary ((opr,ty),x) 	=> Unary ((opr, tySubst subst ty), tySpec subst x)

	| Binary ((opr,ty),x,y) => Binary ((opr, tySubst subst ty), tySpec subst x, tySpec subst y)

	| Quantor ((qtr,ty), vs, e, f) 
		=>
		let
		fun foo (x,ty) = (x, tySubst subst ty)
		in
		Quantor ((qtr, tySubst subst ty), map foo vs, tySpec subst e, tySpec subst f)
		end
(*
	| Quantor ((qtr,ty), vs, e, f) 
		=>
		Quantor ((qtr, tySubst subst ty), vs, tySpec subst e, tySpec subst f)
*)

	| Enum ((enr,ty), es)
		=>
		Enum ((enr, tySubst subst ty), map (tySpec subst) es)

	| Tuple (ex1,ex2) 
		=> 
		Tuple (tySpec subst ex1, tySpec subst ex2)

(* ============================================================

  This function add type info to all identifiers in an expression.
  It does this in the following way:

	(1) If an identifier occurs in the given static context,
	    its type is retrieved from the context, and all type
	    variables there are freshen (subsituted with new/
	    fresh type variables).

	(2) If an identifier does not occur in the static context,
	    but occur in the dynamic context, then its type is
	    retrived from the dynamic context. The type variables
	    occuring in the resulting type should NOT be refreshed.

	(3) If an identifier does not occur in either context, it
	    is given a fresh type variable, and is added to the
	    dynamic context.

Since (3) may extend the dynamic context, the function also returns
the new dynamic context. Care should be taken when handling
quantifiers because of the introduction of bound variables.

============================================================ *)



fun extendContext c_old c_new
	=
	let
	fun p (v,_) = not (MoreList.elem v (map fst c_new))
	in
	List.filter p c_old @ c_new
	end

fun typeIdent		(* retrieve the type of an identifier (or literal) *)
	n		(* fresh type variable counter			*)
	dc		(* list of string,typeL --> dynamic context 	*)
	sc		(* list of string,typeL --> static context 	*)		
	v
	=
	let
	fun is_v (x,_) = (x=v)
	in
	case List.find is_v sc of
	   SOME (_,ty) =>
		let
		val (n',ty') = freshenTyVars n ty
		in
		(n', dc, ty')
		end
	 | NONE => 
		case List.find is_v dc of
		   SOME (_,ty) => (n, dc, ty)
		 | NONE	      => let
				 val a = inventNewVarName n
				 in
				 (n+1, (v,a):: dc, a)
				 end
	end
		 

fun tyPhase1 
	n		(* fresh type variable counter			*)
	dc		(* list of string,typeL --> dynamic context 	*)
	sc		(* list of string,typeL --> static context 	*)		
	expr
	=
	case expr of

	  Lit (_,TyCons ("bool",_))	=> (n, dc, expr)
	| Lit (_,TyCons ("int" ,_))	=> (n, dc, expr)
	| Lit (_,TyCons ("string" ,_))	=> (n, dc, expr)
	| Lit (_,TyCons ("float" ,_))	=> (n, dc, expr)
	| Lit (x,_)			=> 
		let
		val (n',dc',ty') = typeIdent n dc sc x
		in
		(n', dc', Lit (x,ty'))
		end

	| Ident (x,_)			=> 
		let
		val (n',dc',ty') = typeIdent n dc sc x
		in
		(n', dc', Ident (x,ty'))
		end

	| ArrayAccess (a,i)		=> 
		let
		val (na, dca, a') = tyPhase1 n dc sc a
		val (ni, dci, i') = tyPhase1 na dca sc i
		in
		(ni, dci, ArrayAccess (a',i'))
		end

	| App (f,x)			=> 
		let
		val (nf, dcf, f') = tyPhase1 n dc   sc f
		val (nx, dcx, x') = tyPhase1 nf dcf sc x
		in
		(nx, dcx, App (f',x'))
		end

	| Unary ((opr,_),e)		=> 
		let
		val (no, dco, tyo) = typeIdent n  dc  sc opr
		val (ne, dce, e')  = tyPhase1  no dco sc e
		in
		(ne, dce, Unary ((opr,tyo),e'))
		end

	| Binary ((opr,_),e,f)		=> 
		let
		val (no, dco, tyo) = typeIdent n  dc  sc opr
		val (ne, dce, e')  = tyPhase1  no dco sc e
		val (nf, dcf, f')  = tyPhase1  ne dce sc f
		in
		(nf, dcf, Binary ((opr,tyo),e',f'))
		end

	| Quantor ((qtr,_), vs, e,f)	=> 
		let
		val (nq, dcq, tyq) = typeIdent n  dc  sc qtr
		fun opr ((v,_),(n,vs')) = (n+1, (v,inventNewVarName n) :: vs') 
		val (nv,vs') = foldr opr (nq,[]) vs
		val dc' = extendContext dcq vs'
		val (ne, dce, e')  = tyPhase1  nv dc' sc e
		val (nf, dcf, f')  = tyPhase1  ne dce sc f
		val dc_new = extendContext dcf dcq
		in
		(nf, dc_new, Quantor ((qtr,tyq), vs', e', f'))
		end

	| Enum ((enr,_), es)		=> 
		let
		val (nen, dcen, tyen) = typeIdent n dc sc enr
		fun opr (e,(nr,dcr,esr)) = 
			let
			val (ne,dce,e') = tyPhase1 nr dcr sc e
			in
			(ne, dce, e' :: esr)
			end
		val (ns, dcs, es') = foldr opr (nen, dcen, []) es
		in
		(ns, dcs, Enum ((enr,tyen), es'))
		end

	| Tuple (ex1,ex2)	=> 
		let
		val (nf, dcf, ex1') = tyPhase1 n dc   sc ex1
		val (nx, dcx, ex2') = tyPhase1 nf dcf sc ex2
		in
		(nx, dcx, Tuple (ex1',ex2'))
		end


fun tyPhaseOne sc expr = 
	let
	val (n,_,expr') = tyPhase1 0 [] sc expr	
	in
	(n,expr')
	end


fun tiInfer0 n subst expr = (case expr of
	
	  Lit   (v,ty)		=> (n, subst, ty)

	| Ident (v,ty)		=> (n, subst, ty)


	| ArrayAccess (a,i)	=>
		let
		val (na, substa, tya) = tiInfer0 n  subst  a
		val (ni, substi, tyi) = tiInfer0 na substa i
		val tyResult = inventNewVarName ni
		val ty1      = TyCons ("fun", [TyCons ("int",[]), tya])
		val ty2      = TyCons ("fun", [tyi, TyCons ("array", [tyResult]) ])
		val subst'   = mgu (tySubst substi ty1) (tySubst substi ty2)
		
		in
		(ni+1, substi @@ subst', tyResult)
		end

	| App (f,x)		=>
		let
		val (nf, substf, tyf) = tiInfer0 n  subst  f
		val (nx, substx, tyx) = tiInfer0 nf substf x
		val tyResult = inventNewVarName nx
		val tyf2     = TyCons ("fun", [tyx,tyResult])
		val subst' = mgu (tySubst substx tyf) (tySubst substx tyf2)
		in
		(nx+1, substx @@ subst', tyResult)
		end

	| Unary ((opr,ty),e)	=>
		let
		val (ne, subste, tye) = tiInfer0 n subst e
		val tyResult = inventNewVarName ne
		val tyopr'   = TyCons ("fun", [tye,tyResult])
		val subst'   = mgu (tySubst subste ty) (tySubst subste tyopr')
		in
		(ne+1, subste @@ subst', tyResult)
		end

	| Binary ((opr,ty),e,f)	=>
		let
		val (ne, subste, tye) = tiInfer0 n  subst  e
		val (nf, substf, tyf) = tiInfer0 ne subste f
		val tyResult = inventNewVarName nf
		val ty2     = TyCons ("fun", [tye, TyCons ("fun", [tyf, tyResult])])
		val subst' = mgu (tySubst substf ty) (tySubst substf ty2)
		in
		(nf+1, substf @@ subst' , tyResult)
		end


	| Quantor ((qtr,ty),_,e,f) =>
		let
		val (ne, subste, tye) = tiInfer0 n  subst  e
		val (nf, substf, tyf) = tiInfer0 ne subste f
		val tyResult = inventNewVarName nf
		val (n',ty1) = freshenTyVars (nf+1) ty
		val ty2	   = TyCons ("fun", [TyCons ("bool",[]), ty1])
		val ty3    = TyCons ("fun", [tye, TyCons ("fun", [tyf, tyResult])])
		val subst' = mgu (tySubst substf ty2) (tySubst substf ty3)
		in
		(n', substf @@ subst',  tyResult)
		end


	| Enum ((enr,ty),es)	=>
		let
		fun opr (e, (nr,sr,tysr)) = 
			let
			val (n', s', tye') = tiInfer0 nr sr e
			in
			(n', s', tye'::tysr)
			end
		val (n', subst', tys') = foldr opr (n,subst,[]) es
		val tyResult = inventNewVarName n'
		fun f ty = TyCons ("fun", [ty, tyResult])
		val tys''= map f tys'
		val t1	 = TyCons ("dummy", replicate (length tys'') ty)
		val t2   = TyCons ("dummy", tys'')
		val subst'' = mgu (tySubst subst' t1) (tySubst subst' t2)
		in
		(n' + 1, subst' @@ subst'', tyResult)
		end

	| Tuple (ex1,ex2)  =>
		let
		val (nf, substf, tyx1) = tiInfer0 n  subst  ex1
		val (nx, substx, tyx2) = tiInfer0 nf substf ex2
		val tyResult = TyCons ("prod", [tyx1,tyx2])
		in
		(nx, substx, tyResult)
		end
	
	handle TypeLib.unificationFail (tyx1,tyx2) 
	=> 
	raise Fail (failMesg "ExprLib.tyInfer0" expr (TypeLib.failMesg(tyx1,tyx2))))

(* ============================================================
(*DOC

<ITEM>tiInfer
<TYPE>tiInfer : (string * TypeDD.Type) list -> ExprDD.Expr -> (ExprDD.Expr * TypeDD.Type)
<CALL>tiInfer sc expr

<DESC>A static context is a list that associates an identifier or a
literal with its type. This is typically extracted from exConst.sml,
plus information extracted from the variable declaration section of
the target programming language.<p>

Given a static context sc, the function tiInfer will infer the type of
expr, and will also produce a new expression in which all identifiers
and literals have been correctly typed. <p>

Type inference goes in a number of phases. It is easier to explain it
by example. Let say we have an expression:

<pre>(x=y) /\ (!x: hd x = t : hd t = u) ==> (x=1)</pre>

The parser will buid an Expr tree, in which all identifiers have the
type a. Given this tree, tiInfer works as follows:

<OL>

<li> It consults the static context. Identifiers appearing in the
static context will be typed accordingly. However, the type variables
are renamed to be fresh. So the two <tt>hd</tt>'s above will have
different types, e.g. <tt>[e]->e</tt> and <tt>[g]->g</tt>. Notice that
they should indeed have different type. So had we type them equal, say
<tt>[e]->e</tt>, type unification will later fail. <p>

<li>Each identifier which do not occur in the static context is given
the type a', in which a' is fresh. Note that the x in x=y and in x=1
should get the same initial type variable. Whereas the x in the
quantification should get a different type variable (because of the
bounding). <p>

So, after the above two steps, the expression has the following typing
(we will not write down the type of ALL identifiers in the
expression):

<blockquote>
<pre>((x:a)=(y:b)) 
(/\:c->c->bool) 
(!(x:d): (hd:[e]->e) (x:d) = (t:f) : (hd:[g]->g) (t:f) = (u:h)) 
(==>:i->i->bool) 
((x:a)=1)
</pre>
</blockquote>

The above two steps are carried out by the help function
<tt>tyPhaseOne</tt>. <p>

<li>We recursively unify the types in the resulting expression. This
is done by the function <tt>tiInfer0</tt>. The funtion returns a
unifier (type substitution) and the infered type. <p>

<li>The resulting unifier is applied by to the original expression to
correctly type all indentifiers in the expression. This is done by the
function <tt>tySpec</tt>. After this we get the correct type:

<blockquote>
<pre>((x:int)=(y:int)) 
(/\:bool->bool->bool) 
(!(x:[[h]]): (hd:[[h]]->[h]) (x:[[h]]) = (t:[h]) : (hd:[h]->h) (t:[h]) = (u:h)) 
(==>:bool->bool->bool) 
((x:int)=1)
</pre>
</blockquote>
<p>

<li>Finally, the resulting typing is also refreshed with base 0, and get this:

<blockquote>
<pre>((x:int)=(y:int)) 
(/\:bool->bool->bool) 
(!(x:[[a]]): (hd:[[a]]->[a]) (x:[[a]]) = (t:[a]) : (hd:[a]->a) (t:[a]) = (u:a)) 
(==>:bool->bool->bool) 
((x:int)=1)
</pre>
</blockquote>
<p>
</OL>

DOC*)
============================================================ *)


fun tiInfer sc expr 
	= 
	let
	val (n,expr')	  	= tyPhaseOne sc expr
	val (_,subst,ty)  	= tiInfer0 n [] expr' 
	fun snd (_,y)		= y
	val tvars		= (MoreList.nub o List.concat o map (tv o snd)) subst
	val subst_rewind	= freshenTyVars0 0 tvars
	val subst' 		= subst @@ subst_rewind
	in
(*
	(tySpec subst' expr', tySubst subst' ty, subst, subst_rewind, subst')
*)

	(tySpec subst' expr', tySubst subst' ty)

	end


fun tiInferX sc known_expr_ty expr 
	= 
	let
	val (n,expr')	  	= tyPhaseOne sc expr
	val (n',subst,ty)  	= tiInfer0 n [] expr' 
	
	val (_,expr_ty)		= freshenTyVars n' known_expr_ty
	val  subst'		= mgu (tySubst subst ty) expr_ty

	val subst''		= subst @@ subst'

	fun snd (_,y)		= y
	val tvars		= (MoreList.nub o List.concat o map (tv o snd)) subst''
	val subst_rewind	= freshenTyVars0 0 tvars
	val subst_final		= subst'' @@ subst_rewind
	in

	(tySpec subst_final expr', tySubst subst_final known_expr_ty)

	handle TypeLib.unificationFail (tyx1,tyx2) 
	=> 
	raise Fail (failMesg "ExprLib.tiInferX" expr (TypeLib.failMesg(tyx1,tyx2)))
	
	end





(*
fun tiInfer expr 
	= 
	let
	val (_,subst, ty) = tiInfer0 15 [] expr
	in
	(tySpec subst expr, tySubst subst ty)
	end
*)

(* examples:

load "ExprDD" 		;
load "TypeDD"		;
load "TypePara" 	;
load "TypePP" 		;
load "ExprLib"		;
load "TypeLib"		;
open TypeDD TypePP TypePara ExprDD ExprLib TypeLib ;
installPP ppType ;

val ex1 = 
    let
    val x_map   = Ident ("map", parseInLine "(a->b)->[a]->[b]")
    val x_s	= Ident ("s", parseInLine "a")
    in
    App ( App(x_map, x_map), x_s)
    end ;

*)


fun conjureNewName old_name existing_names =
	let
	val x'    = old_name ^ "'"
	val x''   = old_name ^ "''"
	val x'''  = old_name ^ "'''"
	val x_1	  = old_name ^ "_1"
	val x_2	  = old_name ^ "_2"
	val x_3	  = old_name ^ "_3"
	in
	if not (elem x'   existing_names) then x'   else
	if not (elem x''  existing_names) then x''  else
	if not (elem x''' existing_names) then x''' else
	if not (elem x_1  existing_names) then x_1  else
	if not (elem x_2  existing_names) then x_2  else
	if not (elem x_3  existing_names) then x_3  else
		let
		fun leq (s,t) = if s<t then LESS else if s=t then EQUAL else GREATER
		val z = List.last (Listsort.sort leq existing_names)
		in
		z ^ "'"
		end
	end

fun fv0 expr = case expr of

	  Ident (x,_) 		=> [x]
	| Lit _			=> []
	| ArrayAccess (a,i) 	=> fv0 a @ fv0 i
	| App (f,x) 		=> fv0 f @ fv0 x
	| Unary (_,e)		=> fv0 e
	| Binary (_,e,f)	=> fv0 e @ fv0 f	
	| Quantor (_,vs,e,f)	=> 
		let
		fun fst (x,_) = x
		fun p x = (not o elem x o map fst) vs
		in
		List.filter p (fv0 e @ fv0 f)
		end
	| Enum (_,es)		=> List.concat (map fv0 es)
	| Tuple (ex1,ex2)	=> fv0 ex1 @ fv0 ex2 

val fv = nub o fv0


(* 

Rename all free occurence of variable x with y. Renaming is carried
out regardless the type of x. It is therefore assumed that all free
occurence of x has the same type.

*)

fun renameVar x y expr = case expr of
	
	  Ident (z,ty)		=> if z=x then Ident (y,ty) else expr

	| Lit _			=> expr

	| ArrayAccess (a,i)	=> ArrayAccess (renameVar x y a, renameVar x y i)

	| App (f,z)		=> App (renameVar x y f, renameVar x y z)

	| Unary (opr,e)		=> Unary (opr, renameVar x y e)

	| Binary (opr,e,f)	=> Binary (opr, renameVar x y e, renameVar x y f)

	| Quantor (qtr,vs,e,f)	=>

		if elem x (map fst vs) 
		then expr
		else Quantor (qtr,vs, renameVar x y e, renameVar x y f)	

	| Enum (enr, es)	=> Enum (enr, map (renameVar x y) es)

	| Tuple(ex1, ex2)	=> Tuple (renameVar x y ex1, renameVar x y ex2)

(* 

Refresh the name of a given bound variable x in a quantification. That
is, x is renamed with a fresh name. The function also takes a list of
forbidden names --the new name of x is should not conflict with
forbidden names.

*)

fun refreshBound forbidden_names x (expr as (Quantor (qtr,vs,e,f)))
	=
	if not (elem x (map fst vs))

	then expr
	
	else let
	     val used_names = nub (forbidden_names @ map fst vs @ fv e @ fv f)
	     val new_x      = conjureNewName x used_names
	     fun foo (v,ty) = if v=x then (new_x,ty) else (v,ty)
	     val new_vs     = map foo vs
	     val new_e      = renameVar x new_x e
	     val new_f      = renameVar x new_x f
	     in
             Quantor (qtr, new_vs, new_e, new_f)
	     end


fun Subst0 x fv_expr_source expr_source expr_target 
	= 
	case expr_target of

	  Ident (y,_)		=> if x=y then expr_source else expr_target

	| Lit _			=> expr_target

	| ArrayAccess (a,i)	=> ArrayAccess 
					(Subst0 x fv_expr_source expr_source a,
					 Subst0 x fv_expr_source expr_source i		)

	| App (f,y)		=> App 	(Subst0 x fv_expr_source expr_source f,
					 Subst0 x fv_expr_source expr_source y		)


	| Unary (opr,e)		=> Unary (opr, Subst0 x fv_expr_source expr_source e	)

	| Binary (opr,e,f)	=> Binary (opr, 
					Subst0 x fv_expr_source expr_source e,
					Subst0 x fv_expr_source expr_source f		)
	| Quantor (qtr,vs,e,f)	=>
		if elem x (map fst vs)
		then (* x is here a bound var, so no subst. is applied	*)
		     expr_target

		else 
			let

			val conflict = intersect (map fst vs) fv_expr_source

			fun opr (y,exprR) = refreshBound fv_expr_source y exprR	
			
			val Quantor (_,vs',e',f') = foldr opr expr_target conflict

			in
			
			Quantor (qtr, vs', 
				Subst0 x fv_expr_source expr_source e',
				Subst0 x fv_expr_source expr_source f'			)
			end
	
	| Enum (enr,es)		=> Enum (enr, 
					map (Subst0 x fv_expr_source expr_source) es	)
	| Tuple (ex1,ex2)	=> Tuple (Subst0 x fv_expr_source expr_source ex1,
					  Subst0 x fv_expr_source expr_source ex2	)


fun Subst x expr_source expr_target 
	= 
	Subst0 x (fv expr_source) expr_source expr_target 

fun getNmIdent (ExprDD.Ident (nm,ty)) = nm
  | getNmIdent _  = raise Fail "ExprLib.getNmIdent: Invalid use!"

end
end



