structure LabelRules = struct

exception Label_Error of string

local 

open ExprDD TypeDD ExprBiNF 


fun getActualLabelName labs_table symlab 
	=
	case List.find symlab labs_table of
	  SOME (_,al)	=> al
	| NONE		=> raise Label_Error ("User defined label " ^ symlab ^ " cannot be found in the labels table.")



in

(* CONVENTION: An identifier denotes a location (program control)
label of a process P_name if the identifier's name starts with
SYSLAB.<P_name>.    .*)

fun isLabel (Ident (p, TyCons ("bool",[])))
	= 
	String.isPrefix "SYSLAB." p

  | isLabel _					    = false


fun getIdentName (Ident (p,_)) = p

(* The clause c is assumed to be already nubbed, contradictions have
also been removed. So it cannot contain both the positive and negative
form of the same label. *)


fun getPosLabels (poses,_) 
	= 
	(map getIdentName o List.filter isLabel) poses

fun getNegLabels (_,negs)
	= 
	(map getIdentName o List.filter isLabel) negs

(* A clause contains conflicting labels if it contains two distinct
positive labels from the same process.

It is assumed that the parser ensures that expressions only contain
symbolic labels (no internal labels generated by the parser and no
non-existing labels).

It is assumed that the parser ensures that labels occuring in a
process are unique (within the process).

It is assumed that a process can add any number of invisible skips;
hence a clause with a series of conjunction of negative labels never
cause contradiction.

*)

fun getProcessName labelname 
	= 
	let
	fun isDot  a = (a = #".")
	fun notDot a = (a <> #".")
	in

	(String.implode				o
	 MoreList.takeWhile notDot 		o
	 MoreList.dropWhile isDot		o
	 MoreList.dropWhile notDot 		o
	 String.explode				)

	labelname
	end

fun getLabelsOfProcess pname labels
	=
	let
	fun isOfProcess p lab_name = String.isPrefix ("SYSLAB." ^ p ^ ".") lab_name
	in
	List.filter (isOfProcess pname) labels
	end 



fun conflictingLabels labs_table c
	=
	let
	val pos_labs  = getPosLabels c
	val processes = MoreList.nub (map getProcessName pos_labs)
	fun foo p = 
		let
		val pos_labs_of_p = getLabelsOfProcess p pos_labs
		in
		(length						o
		 MoreList.nub					o
		 map (getActualLabelName labs_table)		
		)
  		pos_labs_of_p
		end

	in
	List.exists (fn cnt=> cnt>1) (map foo processes)
	end 

(* A label lab of process P_name is enabled in a clause c if lab does
not occur as a negative label in c and if either of the following hold:

(1) c contains no positive label of process P_name
(2) c contains exactly one label of process P_name, and that is lab

Note that the process name to which lab belong can be obtained from
lab itself. *)


fun isLabelEnabled labs_table lab c
	=
	let
	val p_name	 = getProcessName lab
	val pos_labs 	 = getLabelsOfProcess p_name (getPosLabels c)
	val neg_labs	 = getLabelsOfProcess p_name (getNegLabels c)
	val actual_plabs = map (getActualLabelName labs_table) pos_labs
	val actual_nlabs = map (getActualLabelName labs_table) neg_labs
(*	val lab'         = getActualLabelName labs_table lab *)
	in
	List.all (fn li=> li<>lab) actual_nlabs
	andalso
	(null actual_plabs orelse actual_plabs = [lab])
	end

fun isLabelEnabledClause labs_table lab c
	=
	isLabelEnabled labs_table lab c
	andalso
	not (conflictingLabels labs_table c)

fun takeEnabledClausesOnly labs_table lab nf
	=
	List.filter (isLabelEnabledClause labs_table lab) nf

fun removeLabel (poses,negs)
	=
	(List.filter (not o isLabel) poses, List.filter (not o isLabel) negs)

fun reduceExprOnly  expr
	=
	let
	val (encoded_nf,decodeAtomFunc) = ExprNF.normalize expr
	val nf = ExprNF.decodeNF decodeAtomFunc encoded_nf
	in
		(ExprNF.fromNF o map removeLabel) nf
	end

fun labelReductionRule labs_table lab expr
	=
	let
	val (encoded_nf,decodeAtomFunc) = ExprNF.normalize expr
	val nf = ExprNF.decodeNF decodeAtomFunc encoded_nf
	in
	(ExprNF.fromNF				o
 	 takeEnabledClausesOnly labs_table lab	
	)
	nf
	end

fun labelReductionRule_with_labRemoval labs_table lab expr
	=
	let
	val (encoded_nf,decodeAtomFunc) = ExprNF.normalize expr
	val nf = ExprNF.decodeNF decodeAtomFunc encoded_nf
	in
	(ExprNF.fromNF				o
	 map removeLabel			o
 	 takeEnabledClausesOnly labs_table lab	
	)
	nf
	end



(*

let lab0 be start label of action
    lab1 be end label of action

to prove {p /\ ~q} action {p \/ q}  

--
p CO q = !a. {p} a {q}
--

p = lab0 /\ x>0
q = lab0 /\ x>1

take only lab0-enabled clauses of p /\ ~q; call this (p /\ ~q)-lab0

take only lab1-enabled clauses of p \/ q; call this (p \/ q)-lab1

Prove {(p /\ ~q)-lab0} action {(p \/ q)-lab1} instead

in which the effect of action on labels has been removed.

---

to prove {p /\ ~q} action {p \/ q}  assuming J is stable:

prove {J /\ (p/\~q)-lab0} action {(p\/q)-lab1} instead.

---

to prove {J} action {J}, prove instead

{J-lab0} action {J-lab1}

---

to prove {p /\ ~q} action {q} in the context of ensures:

prove the unless part first. Then we know that:

{p /\ ~q} action {p \/ q} holds. Then it is sufficient to prove:

{(p /\ ~q)-lab0} action {q-lab1}



*)



end
end

