structure PLGALib = struct

exception PLGA_Static_Error of string

local

open TypeDD ExprDD PLSEQDD PLGADD 

fun fst (x,_) = x
fun snd (_,y) = y

val tya		= TyVar "a"
val tybool 	= TyCons ("bool",[])
val tyint	= TyCons ("int",[])
fun tychan a    = TyCons ("chan", [a])
fun tylist a    = TyCons ("list", [a])

fun tyfun a b	= TyCons ("fun", [a,b]) 

val tybool_binop = tyfun tybool (tyfun tybool tybool)

val mk_T		= Lit ("T", tybool)
val mk_F		= Lit ("F", tybool)

fun mk_and (p,q) 	= Binary (("/\\", tybool_binop), p, q)
fun mk_or (p,q) 	= Binary (("\\/", tybool_binop), p, q)
fun mk_not p 		= Unary (("~", tyfun tybool tybool), p)

fun mk_ident x ty	= Ident (x,ty)

(* ~(chanFull >> c) *)

fun mk_guard_send c ty_el 
	=
	let
	val ty1 = tyfun ty_el tybool
	val tyc = tychan ty_el
	in
	mk_not (App (mk_ident "chanFull" ty1, mk_ident c tyc))
	end

(* ~(chanEmpty >> c) *)

fun mk_guard_rec c ty_el 
	=
	let
	val ty1 = tyfun ty_el tybool
	val tyc = tychan ty_el
	in
	mk_not (App (mk_ident "chanEmpty" ty1, mk_ident c tyc))
	end

(* 	"c = send >> c >> e ; SYS.HIST.c = c ++ [e]" *)

fun mk_send c e ty_el
	=
	let
	val ty_c 	= tychan ty_el
	val ty_send 	= tyfun ty_c (tyfun ty_el ty_c)
	val ty_hist	= tylist ty_el
	val ty_append	= tyfun ty_hist (tyfun ty_hist ty_hist)
	val ty_enum	= tyfun ty_el ty_hist

	val sing_e	= Enum (("-[.]-", ty_enum),[e])
	val send_exp 	= App (App (mk_ident "send" ty_send, mk_ident c ty_c), e)
	val c_hist   	= "SYS.HIST." ^ c
	val hist_exp 	= Binary (("++", ty_append), mk_ident c_hist ty_hist, sing_e)

	in
	Seq [Asg (c,send_exp) , Asg (c_hist, hist_exp) ]
	end
	

(* x = fstMsg >> c ; c = flushFstMsg >> c  *)

fun mk_rec c x ty_el
	=
	let
	val ty_c 	= tychan ty_el
	val ty_fstMsg	= tyfun ty_c ty_el
	val ty_flush	= tyfun ty_c ty_c

	val fstMsg	= App (mk_ident "fstMsg" ty_fstMsg, mk_ident c ty_c)
	val flushMsg	= App (mk_ident "flushFstMsg" ty_flush, mk_ident c ty_c)

	in
	Seq [Asg (x, fstMsg), Asg (c, flushMsg)]
	end


fun type_of sc x = case List.find x sc of
	  SOME (_,ty) => ty
	| NONE	      => raise PLGA_Static_Error ("Variable " ^ x ^ " is used but undeclared.")

fun getChanElTy sc c 
	=
	let
	val ty_c = type_of sc c
	in
	case ty_c of 
		(TyCons ("chan",[ty_el]))	=>ty_el
		|  _	=>
			raise PLGA_Static_Error ("Send/Recieve on " ^ c ^ ", which is not of type channel.") 
	end


fun getGuard sc action 
	= 
	let 
	fun worker action = case action of

	  Send ((g,c,_),_) 	=> 
				let
				val ty_el = getChanElTy sc c
				in
				mk_and (g, mk_guard_send c ty_el)
				end

	| Rec  ((g,c,_),_)	=> 
				let
				val ty_el = getChanElTy sc c
				in
				mk_and (g, mk_guard_rec c ty_el)
				end

	| Goto ((g,_),_)	=> g

	| Atomic ((g,_),_)	=> g

	| Block ([],_)		=> mk_T
	| Block ((a ::_),_)	=> worker a

	| Alt ((acs, ElseBlock),_)	=> foldr mk_or mk_F (map worker acs)
	| Alt ((_, ElseExit),_)		=> mk_T
	| Loop ((acs, ElseBlock),_) 	=> foldr mk_or mk_F (map worker acs)
	| Loop ((_, ElseExit),_)	=> mk_T
	
	in
	worker action
	end



in

(* This function will put label info on each atomic action in a given
PLGA body. An action in the body may have a symbolic label which we
get from the PLGA parser. If the parser detects no symbolic label for
an action then a dummy "" (empty string) will be put as the label of
that action.

The function expects a start and end label for the body. Usually, when
called from the top level, it is "SYSLAB.<p_name>.0" and
"SYSLAB.<p_name>.1".

Note that a symbolic label name never begins with a digit. (the parser
will enforce this).

System generated labels all have the form:

	"SYSLAB.<p_name>.<digits-dots-sequence>"

For each action a label info is constructed. This will contains the
symbolic start location label for the action (if present, otherwise
""), its system generated start label, and its system generated end
label.

*)


fun labelize0 startLab endLab plga 
	= 
	case plga of

	  Send (z,symlab) 		=> Send (z, (symlab, startLab, endLab))
	| Rec  (z,symlab) 		=> Rec  (z, (symlab, startLab, endLab))
	| Goto (z,symlab)		=> Goto (z, (symlab, startLab, ""))
	| Atomic (z,symlab)		=> Atomic (z,(symlab, startLab, endLab))
	| Block (acs, symlab)		=>
		let

		fun mk_lab rootLab n = rootLab ^ "." ^ Int.toString n

		fun labelize _ []      = []
		  | labelize n (a::s) = 
			let 
			val sLab = mk_lab startLab n
			val eLab = mk_lab startLab (n+1)
			val a'   = labelize0 sLab eLab a
			in
			a' :: labelize (n+1) s
			end

		val acs' = labelize 0 acs

		in

		Block (acs', (symlab, startLab, endLab))

		end

	| Alt ((acs,else_flag),symlab)
		=>
		let
		val acs' = map (labelize0 startLab endLab) acs
		in
		Alt ((acs', else_flag), (symlab, startLab, endLab))
		end		

	| Loop ((acs,else_flag),symlab)
		=>
		let
		val acs' = map (labelize0 startLab startLab) acs
		in
		Loop ((acs', else_flag), (symlab, startLab, endLab))
		end	
	

fun repBeginLab lab (z,(sl,_,el)) = (z,(sl,lab,el))
fun repEndLab   lab (z,(sl,bl,_)) = (z,(sl,bl,lab))

fun do_RepBeginLab lab action 
	=
	case action of
	  Send x	=> Send   (repBeginLab lab x)
	| Rec  x	=> Rec    (repBeginLab lab x)
	| Goto x	=> Goto   (repBeginLab lab x)
	| Atomic x	=> Atomic (repBeginLab lab x)
	| Block x	=> Block  (repBeginLab lab x)
	| Alt x		=> Alt    (repBeginLab lab x)
	| Loop x	=> Loop   (repBeginLab lab x)
	
fun do_RepEndLab lab action 
	=
	case action of
	  Send x	=> Send   (repEndLab lab x)
	| Rec  x	=> Rec    (repEndLab lab x)
	| Goto x	=> Goto   (repEndLab lab x)
	| Atomic x	=> Atomic (repEndLab lab x)
	| Block x	=> Block  (repEndLab lab x)
	| Alt x		=> Alt    (repEndLab lab x)
	| Loop x	=> Loop   (repEndLab lab x)

fun labelize1 plga
	= 
	case plga of

	  Block (acs,(symlab,bLab,eLab)) =>

		let
		fun foo []     = []
		  | foo (a::s) = do_RepBeginLab bLab a :: s
		
		fun goo []     = []
		  | goo [a]    = [do_RepEndLab eLab a]
		  | goo (a::s) = a :: goo s

		val acs'  = (goo o foo) acs
		val acs'' = map labelize1 acs'
		in
		Block (acs'', (symlab,bLab,eLab))
		end

	| Alt ((acs,flag),(symlab,bLab,eLab)) =>
		let
		fun foo a = do_RepEndLab eLab (do_RepBeginLab bLab a)
		val acs'  = map foo acs
		val acs'' = map labelize1 acs'
		in
		Alt ((acs'',flag), (symlab,bLab,eLab))
		end

	| Loop ((acs,flag),(symlab,bLab,eLab)) =>
		let
		fun foo a = do_RepEndLab bLab (do_RepBeginLab bLab a)
		val acs'  = map foo acs
		val acs'' = map labelize1 acs'
		in
		Loop ((acs'',flag), (symlab,bLab,eLab))
		end

	 | _ => plga
	



fun add_symlab ("",_) 		  tab = tab
  | add_symlab (symlab,actualLab) tab = (symlab, actualLab) :: tab

(* The worker function to construct symbolic label association table.

Ack! Note there can be multiple labels associated with the same
physical program control location.

So... we collect all associations first. This is later used to resolve
GOTO actions.

We assume the parser already guarantees that each user defined
(symbolic) label is unique (at least within each process).

Later, we also need a function to compute 'equivalent labels' of a
given symbolic label lab. This is the list of all symbolic labels
(including lab), that points to the same pysical location as lab.

Boy....
*)



fun mk_label_tab0 tab plga = case plga of

	  Send (_,(sl,al,_))		=> add_symlab (sl,al) tab
	| Rec  (_,(sl,al,_))		=> add_symlab (sl,al) tab
	| Goto 	(_,(sl,al,_))		=> add_symlab (sl,al) tab
	| Atomic (_,(sl,al,_))		=> add_symlab (sl,al) tab
	| Block (acs,(sl,al,_))		=> 
		let
		fun opr (a,tabr) = mk_label_tab0 tabr a
		val tab' = foldr opr tab acs
		in
		add_symlab (sl,al) tab'
		end
	| Alt ((acs,_),(sl,al,_))	=> 
		let
		fun opr (a,tabr) = mk_label_tab0 tabr a
		val tab' = foldr opr tab acs
		in
		add_symlab (sl,al) tab'
		end
	| Loop ((acs,_),(sl,al,_))	=> 
		let
		fun opr (a,tabr) = mk_label_tab0 tabr a
		val tab' = foldr opr tab acs
		in
		add_symlab (sl,al) tab'
		end


fun mk_label_tab plga = mk_label_tab0 [] plga

(* a function to construct the label table ... assume the program name
is "P".*)

fun mk_labels_table preLabel program
	= 
	let
	fun add_pname (sl,al) = (preLabel ^ sl, al)
	in
	    List.map add_pname (mk_label_tab program)
	end ;


exception PLGASyntaxError of string

fun resolve_goto tab plga 
	= 
	let
	fun worker plga = case plga of

	  Send _		=> plga
	| Rec _			=> plga
	| Goto ((g,target),(sl,bl,_)) 
		=>
		(case List.find target tab of

		     NONE     		=> raise PLGASyntaxError ("Target label cannot be found in goto " ^ target)

		    | SOME (_,el) 	=> Goto ((g,target),(sl,bl,el))
		)

	| Atomic _		=> plga

	| Block (acs,labs)	
		=>
		Block (map worker acs, labs)

	| Alt ((acs,flag),labs)
		=> 
		Alt ((map worker acs, flag), labs)

	| Loop ((acs,flag),labs)
		=> 
		Loop ((map worker  acs, flag), labs)

	in
	worker plga
	end


(* Given a list of user defined (symbolic) labels, this function
checks whether each label occurs only once (thus each label is
unique). *)

fun allLabelsAreUnique s =
	let
	val s' = Listsort.sort String.compare s
	fun foo [] 		 	= false
	  | foo [x] 		 	= false
	  | foo (x::(z as (y::_))) 	= (x=y) orelse foo z
	in
	not (foo s')
	end

(* This function generates location labels for a given PLGA body. It
also checks if every symbolic label is unique. *)

fun labelize pname plga =
	let
	val start_label 	= "SYSLAB." ^ pname ^ ".0"
	val end_label 		= "SYSLAB." ^ pname ^ ".1"
	val plga' 		= labelize1 (labelize0 start_label end_label plga)
	val tab 		= mk_label_tab plga'
	val sym_labels_list	=  map fst tab
	in
	if allLabelsAreUnique sym_labels_list
	then resolve_goto tab plga'
	else raise PLGA_Static_Error "There are multiply declared user defined labels."
	end


(* Each atomic actions in PLGA body has the form of:

	start_label: guard->action :end_label

The action part will be translated to PLSEQ statement.

Later on we want to verify a Hoare triple like this (in PLGA):

	{P} lab0: g-->action :lab1 {Q}

this traslates to the following verification conditions in PLSEQ:

	(1)	{P-lab0 /\ g} action {Q-lab1}
	(2)	[P /\ ~g ==> Q]
	(3)	[P-~lab0 ==> Q]

It should be noted that PLGA Hoare triples coming from unless, stable,
and ensures specification all has the form of:
 
	{P /\ ...} lab0: g-->action :lab1 {P \/ Q}

for which the verification conditions (2) and (3) above can be dropped.

*)

fun typeCheck sc plga = 
	let
	fun worker plga = case plga of

 	  Send ((g,c,e),info)		=>
			let
			val tychan_el = getChanElTy sc c 
			val (g',_) = ExprLib.tiInferX sc tybool g
			val (e',_) = ExprLib.tiInferX sc tychan_el e
			in
			Send ((g',c,e'),info)
			end

	| Rec ((g,c,x),info)		=>
			let
			val tychan_el = getChanElTy sc c 
			val ty_x = type_of sc x
			val (g',_) = ExprLib.tiInferX sc tybool g
			in
			if ty_x = tychan_el 
			then Rec ((g',c,x),info)
			else raise PLGA_Static_Error ("The types of channel " ^ c ^ 
						      " and recieving variable " ^ x ^ " do not match.")
			end

	| Goto ((g,target),info)	=>
			let
			val (g',_) = ExprLib.tiInferX sc tybool g
			in
			Goto ((g',target),info)
			end 

	| Atomic ((g,stmt),info)	=>
			let
			val (g',_) = ExprLib.tiInferX sc tybool g
			val stmt'  = PLSEQLib.typeCheck sc stmt
			in
			Atomic ((g',stmt'),info)
			end

	| Block (acs, info)		=> Block (map worker acs, info)

	| Alt ((acs,flag), info)	=> Alt ((map worker acs, flag), info)

	| Loop ((acs,flag), info)	=> Loop ((map worker acs, flag), info)


	in
	worker plga
	end
	

(* This function type check a plga body, and generates location
labels. *)

fun preprocess pname sc plga = typeCheck sc (labelize pname plga)
	

(* This function computes the list of all atomic actions in a guven
PLGA body. It assumed that the body has already been type checked. *)


fun spitAtomics sc plga 
	= 
	let
	fun worker plga = case plga of

	  Send ((g,ch,e),linfo)
		=> 
		let
		val ty_el = getChanElTy sc ch 
		val g'	  = mk_and (g, mk_guard_send ch ty_el)
		in
		[(mk_send ch e ty_el, g', linfo)]
		end

	| Rec ((g,ch,x),linfo)
		=>
		let
		val ty_el = getChanElTy sc ch 
		val g'	  = mk_and (g, mk_guard_rec ch ty_el)
		in
		[(mk_rec ch x ty_el, g', linfo)]
		end

	| Goto ((g,_),linfo)		=> [(Skip,g,linfo)]
	| Atomic ((g,stmt),linfo)	=> [(stmt,g,linfo)]
	| Block (acs,_)			=> List.concat (map worker acs)

	| Alt ((acs,ElseBlock),_)	=> List.concat (map worker acs)
	| Alt ((acs,ElseExit),linfo)	
			=>
			let
			val acs_guards = map (getGuard sc) acs
			val else_guard = mk_not (foldr mk_or mk_F acs_guards)
			in
			(Skip, else_guard, linfo) 
			::
			List.concat (map worker acs)
			end

	| Loop ((acs,ElseBlock),_)	=> List.concat (map worker acs)
	| Loop ((acs,ElseExit),linfo)	
			=>
			let
			val acs_guards = map (getGuard sc) acs
			val else_guard = mk_not (foldr mk_or mk_F acs_guards)
			in
			(Skip, else_guard, linfo) 
			::
			List.concat (map worker acs)
			end
	
	in
	worker plga
	end


end
end

