(* ----------------------------------------------------------------------
   FileName : PLPCDD.sml
   Structure: PLPCDD
   
   Modified from : 
   File        : PL1astx.sml
   Description : Defining an abstract representation of a UNITY-like
                 programming language PL1. Also defines a corresponding
                 pretty printer.
   Author       : Wishnu Prasetya (wishnu@cs.uu.nl)
   Version      : May 2000
   ---------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------
   ---------------------------------------------------------------------- *)

(* note: 

   Terminology :

   Given a Process,
   > All sibling processes will run when the main process executed.
     The sibling process declaration is not part of Main process
     body programs.
   > Shared Variable. Every variable declare in the argument is 
     shared between the process and its caller,
     although the name is different, (alpha equivalent).
     and this process share relation is transitif and symmetry 
     (not the calling relation).
     Some variable constraint is propagated, 
     base on the constraint rules and the calling relation. 

   > Variable declaration has extra identifier, giving constraint
     on read. There are :
              ROBP = Read only by Parent
              ROBC = Read only by Children  
              (The constraints will apply to all the nested Sibling
               for the RobS, and to all the nested Caller for RobC, 
               which has the variable) 
              ROBM = Read only by Me (this process)
              (This constraint only apply in the process where the constraint
               is given)
     This constrain is guaranteed that it would not raise any conflict.
     But indeed there is a possibility that none can write to the
     variable. In this situation, the variable act like a constant.
     with the value as its initialization value.   
     If no initialization given, no specific treatment will be given.

     Special case :
     If a process with ROBP constraint in one of its argument,
     was called with a constant ? than that constant
     is considered as it initialization. and the variable in 
     the callee process remain invisible in the caller process.

     There is no keyword to abort those constraint,
     it's required to simplify the verification. 



*)

structure PLPCDD = struct

local open TypeDD PLGADD 
in


datatype VarConstr = ROBM   (* Read only by Me (this process)          *) 
                   | ROBC   (* Read only by Children                   *) 
                   | ROBP   (* Read only by Parent                     *)
                   | NONE   
;

datatype VarDecl =
	  NOINIT of 
          VarConstr list  *
	  TypeDD.Type * 
	  string list		(* vars declarations, without initialization     *)
	| INIT of	
          VarConstr list *	
	  TypeDD.Type *
	  string *		(* the name of the variable                      *)
	  ExprDD.Expr 		(* using an expression to initialize it          *)

type VarDeclList = 		(* list of variables declarations ---------------*)
	VarDecl list

datatype ProcSibling =		(* process-instance declaration -----------------*)	
   SINGLE of
   	string *		(* process-instance name                         *) 
	string *		(* process-class                                 *)
	ExprDD.Expr list	(* process-class actual parameters, linking
				   formal parameters to actual shared variables  *)
 | ARRAY of
   	string *		(* process-instance name                         *) 
	string *		(* process-class                                 *)
        ExprDD.Expr *           (* number of process array                       *)
        string *                (* index variable name                           *)
	ExprDD.Expr list	(* process-class actual parameters, linking
	                           formal parameters to actual shared variables  *)

type ProcSiblingList =		(* list of process instances --------------------*)
	ProcSibling list

type ProcClass = 		(* process-class declaration --------------------*)
	string *		(* process-class name                            *)
	VarDeclList *		(* external shared variables, listed as formal 
				   parameters;
				   no initialization is allowed here		 *)
	VarDeclList *		(* main's private variable declarations          *)
	ProcSiblingList *	(* creation of child processes		         *)
	string PLGADD.PLGA_Stmt (* list of statements (the Main)                 *)

type Operator     = (string*TypeDD.Type)

datatype ProcSpec = 
         VALID     of ExprDD.Expr list*ExprDD.Expr
       | STABLE    of ExprDD.Expr list*ExprDD.Expr
       | UNITYREL  of ExprDD.Expr list*Operator*ExprDD.Expr*ExprDD.Expr*ExprDD.Identifier list
       | DNF       of ExprDD.Expr list
       | PRESERVES of ExprDD.Expr list*ExprDD.Expr*ExprDD.Expr*ExprDD.Identifier list*ExprDD.Expr
                      (* J |-- Q at p PRESERVES V UNLESS q  *)
       | GENERIC   of ExprDD.Expr list
(*      not used since xMech 0.4.   Jan 22, 2002 *) 
	| INVARIANT of ExprDD.Expr * ExprDD.Expr    
	(* syntax example invariant J : x>1 *)


(* every spec identified by string *)

type SpecList = (string*ProcSpec) list ;

end
end
