structure PLPCManager  = struct

local 

open HolKernel  Dsyntax Psyntax TypeDD MoreList ;

open String intLib;

in 

(* Initialitation variable global *)

type notes           = string      (* later it would be more complicated *)

val cPath            = ref "";
val cName            = ref "";
val specList         = ref [];
(* val provenList       = ref [];  *)
val currentSpec      = ref Option.NONE;
val provenFlag       = ref false;

(* temporal global variable needed for the verification system *)
val procRep          = ref Option.NONE;    
val specRep          = ref [];    
val programLab       = ref Option.NONE;   
val programRep       = ref Option.NONE;   

val labs_table       = ref [];
val atomics          = ref [];

val totalContext     = ref [];
val procContext	     = ref [];
val extraContext     = ref []; 


(* shortcut function for manipulating global variable *) 
fun getVal (x:'a Option.option ref) = 
	 if (Option.isSome (!x)) 
	 then Option.valOf (!x) 
	 else raise Fail "Error: Option.NONE, improper use !! "
; 

fun putVal x = Option.SOME  x;

(* preparing base static context *)

val level1_StaticContext =
	let
	fun func (v,t,_,_) = (v, TypePara.parseInLine t)
	in 
	    List.map func exConst.context
	end ;


(* Defining interface function *)

(* 
The function below will parse a PLGA code. It will do action
labelling, type checking, and type labelling.
*)

fun parseProgram extraContext fname 
	=
	let
        val total_typeContext = extraContext @ level1_StaticContext
	val prog              = PLGAPara.parse fname
	(* now labelize and type check prog -- use P as process name *)
	val prog'             = PLGALib.preprocess "P" total_typeContext prog
	in
	prog'
	end ;

(*
The function below will parse a PLPC spec file and then type check and
type label every spec loaded from the file.
*)

fun parseSpec extraContext fname
	=
	let 
        val total_typeContext = extraContext @ level1_StaticContext
	val specs             = PLPCPara.parseS fname
	fun typeChecker (sname,spec) 
                              = (sname, PLPCLib.specTypeChecker total_typeContext spec)
	in
	List.map typeChecker specs
	end ;



(* Transformation to HOL,
   Reduce and proving in HOL *) 

(* 
The result of the function above is the passed to the function
below. It will apply the Hoare Logic from PLSEQ to reduce the Hoare
triples to predicates. It then translate them to HOL (syntactical
translation) and then put them in a big conjunction and disjunction,
ready to be set up as a HOL goal.
*)


(* modified extended with assumption list,
   Notes:
   1. Hoare triple List also contain assumption list,
   2. Hoare triple List might not contain spec in hoare triple form.
      (valid specification is in proportional form)   
 *)

fun reduceHOA_and_trans2HOL context (asml,hoare_triple_list1, hoare_triple_list2) 
	= 
	let
	fun conjunct_them s   = list_mk_conj s
	fun disjunct_them s   = list_mk_disj s
	fun reduce hoa        = PLSEQRules.topReduce hoa
	fun toHOL  goals      = list_mk_conj (List.map ExprHol.exprInHol goals)
	fun typeCheck goals   = List.map (fst o (ExprLib.tiInfer context)) goals
	val conjunctive_goals = List.map (toHOL o typeCheck o reduce) hoare_triple_list1
	val disjunctive_goals = List.map (toHOL o typeCheck o reduce) hoare_triple_list2
        val asmlTerm          = List.map ExprHol.exprInHol asml
        fun combine cp dp     = mk_conj (cp, dp)

	in
	    if null hoare_triple_list2 
	    then (asmlTerm, conjunct_them conjunctive_goals)
	    else (asmlTerm, mk_conj  (conjunct_them conjunctive_goals, 
                                      disjunct_them disjunctive_goals))
	end ;

local 
fun checkgoal() = GoalstackPure.drop (goalstackLib.status());
in
fun dropAll() = 
    (goalstackLib.drop();checkgoal();dropAll())
    handle _ => print "\nGoals stack has been initialized.\n"
end    

(* this function will : 
   1. generate the verification condition  of the current spec
   2. Reduce the hoare triple form
   3. transform the the reduced hoare triple into HOL term.

*) 

fun sendToHol curSpec = 
    let val (nm:string, spec) = curSpec

        val (asml_,hoaListConj,hoaListDisj)  
                          = PLPCRules.VCGspec (!labs_table) (!atomics) spec
        val asml          = PLPCLib.asmlPreprocessing (!specRep) asml_
        val asl_goal      = reduceHOA_and_trans2HOL (!totalContext) (asml,hoaListConj, hoaListDisj)
        
    in
        (goalstackLib.set_goal asl_goal;
	 print ("\nSpecification: "^nm^"\n"))
    end

fun useHol () = sendToHol (getVal currentSpec)

(* function to use hol for decomposing composed specification *)


fun sendUnity  (compSpec:PLPCDD.ProcSpec) = 
    if (PLPCLib.isUnitySpec compSpec)  
      then PLPCHol.unityInHol true (!specRep) compSpec
      else raise Fail "PLPCManager.sendUnity: Not a unity specification."

fun decomposeSpec()=
    let val (nm,compSpec) = getVal currentSpec
    in
        if (PLPCRules.compDetect (!specRep) compSpec) 
        then (sendUnity compSpec;
	      print ("Specification: "^ nm ^ "has been send to hol."))
	else  print "No composition information found."      
    end

fun specInHol()=
    let val (nm,compSpec) = getVal currentSpec
    in
        if (PLPCRules.compDetect (!specRep) compSpec) 
        then sendUnity compSpec	     
	else  raise Fail "No composition information found."      
    end

(* getGoal : goalStack -> procSpec List *)
(* fun getGoal () = ??? *)



(* function to identify whether it was succeed or not, for each specification
   Still required human interference  *)
fun confirmProven() =
    (print "Looking for proven goal.... \n"   ;
     Parse.print_thm (goalstackLib.top_thm()) ;
     goalstackLib.drop()                      ;
     print "\nRemoving proven spec...."	      ;
     specList   := List.tl (!specList)        ; 
     provenFlag := true                       ;
     print ("\nSpecification left : "
	    ^ (Int.toString (List.length (!specList)))))

     handle _ => print "no proven goal found"

(* get the new specification *)
fun newSpec() = 
   if (List.null (!specList))
   then (if (!provenFlag) 
         then print "The last specification has been proven.\n" 
         else print "No open specification.\n")
   else if (!provenFlag) 
        then(  
             currentSpec := putVal (List.hd (!specList)) ; 
             provenFlag   := false              )
        else print 
   "The previous specification has not been proven yet, use chooseSpec [name] instead\n"

fun show() = snd (getVal currentSpec)
fun showName() = fst (getVal currentSpec)   

(* something like rotate in hol, but here we use specification name as the pointer *)
fun chooseSpec (nm:string) = 
    let val tempList = !specList 
        fun rotate flag nm [] prevList             
			= if (flag) then prevList
				    else raise Fail ("PLPCManager.chooseSpec: Cannot find "^nm)
							
          | rotate flag nm ((nms,sp)::sl) prevList 
			= if (nm = nms)
                          then (nms,sp)::(prevList @ sl)
                          else rotate flag nm sl (prevList @[(nms,sp)])
    in 
        (specList := rotate false nm tempList [];
	 currentSpec := putVal (List.hd (!specList)))
    end
;

fun initSpec() = 
    (currentSpec := putVal (List.hd (!specList));
     showName())
    

(* Main function the set up the specific environment for each process class *)
fun loadProcess path Name  =  
    let 
        val sName     = path^Name^".spc"
        val pName     = path^Name^".pc"
        val preLabel  = "SYSLAB."^Name^"."
        
    in        
       (
	cPath          := path                                      ; 
	cName          := Name                                      ;
        procRep        := putVal (PLPCPara.parse pName)             ;
 	programRep     := putVal (PLPCLib.getPLGA (getVal procRep)) ;

        specRep        := PLPCPara.parseS sName                     ; 

        procContext    := PLPCLib.getContextInfo (getVal procRep)   ;
        totalContext   := (!extraContext) @ (!procContext) @ level1_StaticContext       ;

 	programLab     := putVal (PLGALib.preprocess (!cName) (!totalContext) (getVal programRep));
        specList       := PLPCLib.specListTypeCheck (!totalContext) (!specRep) ;
        specRep        := !specList ;	

        labs_table     := PLGALib.mk_labels_table preLabel (getVal programLab)   ;
	atomics        := PLGALib.spitAtomics (!totalContext) (getVal programLab)   
        
       )
     end;

fun reload () = loadProcess (!cPath) (!cName) ;


end (* of local  *)
end (* of struct *) 

(* debugging...
 

	(List.map (toHOL o typeCheck o reduce) hoare_triple_list1;print "\n")
handle HOL_expr_err goals => exl:= goals;


*)



