structure PLSEQDD = struct

(* ============================================================
(*DOC

<MODULE>PLSEQDD
<AUTHOR>underdarkPrime
<ADRESS>underdarkPrime@yahoo.com
<DATE>august 2001

<DESC> This module provides the abstract representation of PLSEQ
programs and specifications. <b>PLSEQ</b> is an abbreviation of
<i>Programming Logic for Sequantial programs</i>. It is a very simple
logic. <p>

The language of PLSEQ consists of the language of programs, which is
used to express or write a program, and a specification language which
is used to express a specification of a program. The programming
language is C-like. <p>

The specification language is a simple Hoare triple language. <p>

Both the programming and specification languages share an expression
language. We will used the expression language defined by ExprDD and
ExprParser. It is a very rich expression language, and with it we can
easily express complicated specifications (think for example a sorting
program; the program itself may only use an array, but its
specification is complicated). <p>

To be more precise, PLSEQ is actually not a logic for a complete
programs. It is a logic of sequentially composed statements. It is
intended to be used as a part of a bigger logic, such as PLTEM. PLSEQ
has no language construct to form a complete program. Instead, it is
assumed that PLSEQ specifications are produces by an external tool
that parses a complete program of some other language. <p>

<ITEM>Stmt

<TYPE>datatype Stmt =
     Skip
   | Asg	of string * Expr
   | Inc	of string
   | Dec	of string
   | Seq	of Stmt list
   | IfThen	of Expr * Stmt 
   | IfElse	of Expr * Stmt * Stmt
   | Case	of (Expr * Stmt) list * Stmt
   | While	of Expr * Expr * Stmt
   | ForLoop	of Expr * string * Expr * Expr * Stmt * Stmt

<DESC>

<UL>
<li> <tt>Asg (x,e)</tt> represents assignment <tt>x=e</tt>.<p></li>

<li> <tt>Inc i</tt> represents increment statement <tt>i++</tt>. 
     Conversely <tt>Dec i</tt> represents <tt>i--</tt>.<p></li>

<li> <tt>If (g,s,t)</tt> represents an if-then-else statement. <p></li>

<li> <tt>Case s e</tt> represents a case-statement. The s is a list of
guard the its corresponding statement. The e is the else part. The
else part is compulsory. <p></li>

<li> <tt>While (i,g,s)</tt> represents a while statement. The i is the
loop invariant. <p></li>

<li> <tt>ForLoop (I,i,lb,up,inc,body)</tt> represents a for-loop, for example:
<blockquote>
<pre>inv I
for (i=0; 10; i++) s = s + a[i]</pre>
</blockquote>
The i is called the looping counter. The loop terminates if the value
of i at least equal to the loop upper bound, in this case 10. <p>

A for-loop is restricted in the following way. The loop body should
not contain any assignment to this counter, nor any program variable
occuring in the loop lower bound (lb) and upper bound (ub). This
greatly simplify the termination proof of the loop. We only need to
prove that the loop increases the counter. <p>
</li>

</UL>

DOC*)
============================================================ *)

local 

open ExprDD TypeDD

in

datatype Stmt =
     Skip
   | Asg	of string * Expr
   | Inc	of string
   | Dec	of string
   | Seq	of Stmt list
   | IfThen	of Expr * Stmt 
   | IfElse	of Expr * Stmt * Stmt
   | Case	of (Expr * Stmt) list * Stmt
   | While	of Expr * Expr * Stmt
   | ForLoop	of Expr * string * Expr * Expr * Stmt * Stmt
(* added 22 October *)
   | Relation   of string * Expr
(*   x from relX    *)

(* Examples   (6,2,3,0) (9,7,1,2) (7,1,7,0) 
   Relation of devided number, 6 devided by 2 is 3 remainder 0 
  
   x is number(v,w,x,y) 

   ExprDD.Relasi (("number",`:num->num->num->num->bool`),
                  ExprDD.Ident ("x",`:num`),
                  [ExprDD.Ident ("v",`:num`),
                   ExprDD.Ident ("w",`:num`),
                   ExprDD.Ident ("x",`:num`),
                   ExprDD.Ident ("y",`:num`)])
 
   Relation mesgA is channelContain(chan,mesgA)

   ExprDD.Relasi (("channelContain",`:channel->mesg`),
                  ExprDD.Ident ("mesgA",`:mesg`),
                  [ExprDD.Ident ("chan",`:channel`),
                   ExprDD.Ident ("mesgA",`:mesg`)])

  x is relasiName ( varList ) 

  The informal semantic is :
  if the relasi with relasiName of varList does exist 
  (while the x is as one of its component) then
  the value of x is assign to one of its possible values.

  refer to the first example 

  x from number(6,x,y,0) 
  
  then x is 2 or 3 or 6 or 1,... 
  
  


*)

(* ============================================================ 
(*DOC

<ITEM>SEQSpec
<TYPE>datatype SEQSpec = Hoa of Expr * Stmt * Expr 

DOC*)
============================================================ *)

datatype SEQSpec =
     Every 	of Expr 
   | Hoa   	of Expr * Stmt * Expr 
   | Disj       of SEQSpec list * SEQSpec list

end


end


