structure TypeLib :> TypeLib  = struct

(* ============================================================
(*DOC

<MODULE> TypeLib
<AUTHOR> underdarkPrime
<ADDRESS> underdarkPrime@yahoo.com
<DATE> August 2001

<DESC> Provides functions for type unification and type substitution.

<p>
<b> Terminologies: </b> <p>

A type substitution list is a list of pairs (a,ty) where a is a string
representing a type variable and ty is a type. When applied to a
another type tx, it will substitute all occurences of a in tx with ty.

DOC*)
============================================================ *)

local

open TypeDD MoreList

in


(* to String function *)

(* ====================================================
(*DOC

<ITEM>toString
<TYPE>toString : TypeDD.Type -> string

<DESC>It generates the string representation of type <tt>TypeDD.Type</tt>.

DOC*)
==================================================== *)

fun toString ty = PP.pp_to_string 20 TypePP.ppType ty; 


(* ============================================================
(*DOC

<ITEM>tv
<TYPE>tv : Type -> string list
<CALL>tv ty

<DESC>It returns the list of all type variables occuring in
ty. Duplicates are removed.

DOC*)
============================================================ *)

fun tv0 (TyVar a) = [a]
  | tv0 (TyCons (c,ts)) = List.concat (map tv0 ts)

fun tv ty = nub (tv0 ty)

(* ============================================================
(*DOC

<ITEM> tySubst
<TYPE> tySubst : (string * Type) list -> Type -> Type
<CALL> tySubst subst ty

<DESC> It applies the type substitution list <tt>subst</tt> to the
type <tt>ty</tt>.

DOC*)
============================================================ *)

fun tySubst [] f = f
  | tySubst z  f = case f of
	  TyVar a	=> let
			   fun p (v,_) = a=v
			   in
			   case  List.find p z of
			     NONE 	=> TyVar a
                           | SOME (_,e) => e			   
			   end
	| TyCons (c,ts)	=> TyCons (c, map (tySubst z) ts)


(* ============================================================
(*DOC

<ITEM> @@
<TYPE> @@ : ((string * Type) list * (string * Type) list) -> (string * Type) list
<CALL> s @@ t 

<DESC>Given two type substitution s and t, the operator @@ combines
them in such a way that:

<pre>tySubst (s @@ t) ty = tySubst t (tySubst s ty)</pre>

In otherwords, when the subsitution <tt>s @@ t</tt> is applied to a
type ty, the substitution <tt>s</tt> is applied first, then
<tt>t</tt>. <p>


DOC*)
============================================================ *)

infix @@

fun tsubs1 @@ tsubs2 
	=
	let
	val tsubs1' = map (fn (v,t) => (v, tySubst tsubs2 t)) tsubs1
	in
	tsubs1' @ tsubs2
	end

(* ============================================================
(*DOC

<ITEM> mgu
<TYPE> mgu : Type -> Type -> (string * Type) list
<CALL> mgu ty1 ty2

<DESC>It produces a type substitution that will unifiy ty1 and
ty2. That is, if s is the resulting substitution, we have:

<pre>tySubst s ty1 = tySubst s ty2</pre>

Moreover, s is (suposedly) the most general unifier of ty1 and
ty2. Keep in mind that a pair of types like <tt>a</tt> and
<tt>[a]</tt> cannot be unified. 

<p>If no unifier can be found, an exception will be raised.

<ITEM> unificationFail
<TYPE> exception unificationFail of Type * Type 
<CALL> raise unificationFail (ty1,ty2)

<DESC> The exception is raised when ty1 and ty2 cannot be unified.

DOC*)
============================================================ *)

exception unificationFail of Type * Type 

fun failMesg (ty1,ty2) = ("X-Mech type clash: expression of type:   "^ toString ty1 ^ 
			  "   cannot have type:   "^ toString ty2)


fun mgu ty1 ty2 
	= 
	case (ty1, ty2) of

    (TyVar a, TyVar b)	=> if a=b 
				then [] 
				else if a<b then [(b, TyVar a)] else [(a, TyVar b)]

  | (TyVar a,u2) 	=> if elem a (tv u2)
				then raise unificationFail (ty1,ty2)
				else [(a,u2)]

  | (u1, TyVar b) 	=> if elem b (tv u1)
				then raise unificationFail (ty1,ty2)
				else [(b,u1)]

  | (TyCons (c,ts), TyCons (d,us))
	 =>
	if c<>d then raise unificationFail (ty1, ty2)
		else 
		let
		fun opr ((t1,t2),subst) = 
			let
			val subst' = mgu (tySubst subst t1) (tySubst subst t2)
			in
			subst @@ subst'
			end
		in
		foldr opr [] (ListPair.zip ( ts,us))
		end

(*--

some test function and data

fun tyUnify ty1 ty2 
	= 
	(tySubst (mgu ty1 ty2) ty1
	handle unificationFail (tyx1,tyx2) => raise Fail (failMesg(tyx1,tyx2)))

val ex1 = TyCons ("fun", [TyCons ("list", [TyVar "a"]),
                          TyVar "b"])

val ex2 = TyCons ("fun", [TyVar "c", TyCons ("bool", [])])

val ex3 = TyCons ("fun", [TyCons ("list", [TyCons ("list", [TyVar "d"])]),
                          ex2])

val ex4 = TyCons ("fun", [TyVar "a", TyVar "a"])

val ex5 = TyCons ("fun", [TyCons ("list",[TyVar "b"]), 
                          TyCons ("list",[TyCons ("bool",[])])])

*)




end

end

