structure TypePara :> TypePara = struct	(* Para = Parser access *)

local open TypeDD in

(* ====================================================
(*DOC

<MODULE> TypePara
<AUTHOR> underdarkPrime
<ADDRESS> underdarkPrime@yahoo.com

<DESC> Provide a parser to parse a type expression. Grammar:

<UL>
<li> <tt> int </tt>, <tt> bool </tt>, ... basic types. </li>
<li> Other identifiers : type variables. </li>
<li> <tt> (t,u,v,...) </tt> : tuple (product) type. </li>
<li> <tt> [t] </tt> : list over t. </li>
<li> <tt> {t} </tt> : set over t. </li>
<li> <tt> t->u </tt> : a function type from <tt>t</tt> to <tt>u</tt>.
</UL>

<ITEM>parseInLine
<TYPE>parseInLine : string -> TypeDD.Type

<DESC>It parses the input string and produces the representing value
of type <tt>TypeDD.Type</tt>.

DOC*)
==================================================== *)

fun parseTypePlain lexbuf =
    let val expr = TypeParser.START TypeLexer.Token lexbuf
    in
	Parsing.clearParser();
	expr
    end
    handle exn => (Parsing.clearParser(); raise exn);

(* Parse; show offending program piece on error *)

fun parseTypeReport file stream lexbuf =
    let val expr = 
	    TypeParser.START TypeLexer.Token lexbuf
	    handle
	       Parsing.ParseError f =>
		   let val pos1 = Lexing.getLexemeStart lexbuf
		       val pos2 = Lexing.getLexemeEnd lexbuf
		   in
		       Location.errMsg (file, stream, lexbuf) 
		                       (Location.Loc(pos1, pos2))
		                       "Syntax error."
		   end
	     | TypeLexer.LexicalError(msg, pos1, pos2) =>
		   if pos1 >= 0 andalso pos2 >= 0 then
		       Location.errMsg (file, stream, lexbuf)
		                       (Location.Loc(pos1, pos2))
		                       ("Lexical error: " ^ msg)
		   else 
		       (Location.errPrompt ("Lexical error: " ^ msg ^ "\n\n");
			raise Fail "Lexical error");
    in
	Parsing.clearParser();
	expr
    end
    handle exn => (Parsing.clearParser(); raise exn);

(* Create lexer from instream *)

fun createLexerStream (is : BasicIO.instream) =
  Lexing.createLexer (fn buff => fn n => Nonstdio.buff_input is buff 0 n)

(* Parse an expression from a file *)

(*
fun parse file =
    let val is     = Nonstdio.open_in_bin file
        val lexbuf = createLexerStream is
	val expr   = parseTypeReport file is lexbuf
	             handle exn => (BasicIO.close_in is; raise exn)
    in 
        BasicIO.close_in is;
	expr
    end
*)

(* Parse an expression in line *)

fun parseInLine str = parseTypePlain (Lexing.createLexerString str)

end
end

